import pandas as pd

from sklearn.model_selection import train_test_split

from transformers import AdamW, XLNetTokenizer, TFXLNetModel
import numpy as np
import os
from tensorflow.keras.preprocessing.sequence import pad_sequences

import time
import datetime

import pickle

from ast import literal_eval

from tqdm import tqdm

import tensorflow as tf

from tensorflow.keras.models import Model, Sequential
from tensorflow.keras.layers import Attention, Input, Dense, LSTM, Embedding, Bidirectional, Dropout, TimeDistributed, concatenate, MaxPooling1D, Activation, Add, Flatten, Conv1D, BatchNormalization
from tensorflow.keras.optimizers import Adam

import tensorflow_addons as tfa

def get_tokens(tokenizer, toks1, toks2, cls, sep):
    toks1 = tokenizer.convert_tokens_to_ids(toks1) + [sep]
    toks2 = tokenizer.convert_tokens_to_ids(toks2) + [sep] + [cls]

    indexed_tokens = toks1 + toks2
    segments_ids = [0] * len(toks1) + [1] * len(toks2)
    return indexed_tokens, segments_ids

def createModel(xlnetModel, conv1D_shared, batchN, activa, attent, input_ids0, attention_mask0):

    xlnetout = xlnetModel.transformer({"input_ids": input_ids0, "attention_mask": attention_mask0})
    x = conv1D_shared(xlnetout.last_hidden_state)
    x = batchN(x)
    x = activa(x)
    x = attent([x, x])

    model = Model(inputs=[input_ids0, attention_mask0,], outputs=x)
    return model

def concat_model():
    input_ids0 = Input(shape=(256,), dtype='int64')
    attention_mask = Input(shape=(256,), dtype='int64')

    input_ids1 = Input(shape=(256,), dtype='int64')

    print('Loading XLNetModel')
    xlnetModel = TFXLNetModel.from_pretrained('xlnet-base-cased')
    conv1D_shared = Conv1D(64, kernel_size=(7), strides=(2))
    batchN = BatchNormalization()
    activa = Activation('relu')
    attent = Attention(use_scale=True)

    model1 = createModel(xlnetModel, conv1D_shared, batchN, activa, attent, input_ids0, attention_mask)
    model2 = createModel(xlnetModel, conv1D_shared, batchN, activa, attent, input_ids1, attention_mask)

    combined = concatenate([model1.output, model2.output])

    z = MaxPooling1D((3), strides=(2))(combined)
    # x = Dense(1, activation="relu")(x)
    z = Flatten()(z)
    z = Dense(1, activation="sigmoid")(z)


    model = Model(inputs=[input_ids0, input_ids1, attention_mask,], outputs=z)
    model.summary()
    adam = Adam(learning_rate=0.00001)
    model.compile(optimizer=adam, loss=tfa.losses.ContrastiveLoss(), metrics=['acc', tf.keras.metrics.Precision(), tf.keras.metrics.Recall()])

    return model

if __name__ == '__main__':
    # Use appropriate locations
    training_data_loc = './train_sequences_tokenized12.csv'
    output_loc = './model/trained_model12/'
    # Use appropriate locations
    print('reading data')
    try:
        print('Loading tokenizer')

        tokenizer=XLNetTokenizer.from_pretrained('xlnet-base-cased')

        cls, sep = tokenizer.convert_tokens_to_ids(["<cls>", "<sep>"])

        with open(os.path.join(output_loc, 'input_tokens.pkl'), 'rb') as f:
            input_tokens = pickle.load(f)
        with open(os.path.join(output_loc, 'input_seg_ids.pkl'), 'rb') as f:
            input_seg_ids = pickle.load(f)
        with open(os.path.join(output_loc, 'labels.pkl'), 'rb') as f:
            labels = pickle.load(f)

    except:
        print('Reading training data file...')

        df = pd.read_csv(training_data_loc, usecols=['para1_tokens', 'para2_tokens', 'para1_len', 'para2_len', 'label'])

        df = df[(df['para1_len'] > 0) & (df['para2_len'] > 0)]

        df['para1_tokens'] = df['para1_tokens'].apply(literal_eval)
        df['para2_tokens'] = df['para2_tokens'].apply(literal_eval)



        print('Generating training input...')
        input_tokens = list()
        input_seg_ids = list()
        labels = list()
        for idx, row in tqdm(df.iterrows(), total=df.shape[0]):
            indexed_tokens, segments_ids = get_tokens(tokenizer, row['para1_tokens'], row['para2_tokens'], cls, sep)
            input_tokens.append(indexed_tokens)
            input_seg_ids.append(segments_ids)
            labels.append(row['label'])
        # Use appropriate locations
        with open(os.path.join(output_loc, 'input_tokens.pkl'), 'wb') as f:
            pickle.dump(input_tokens, f)
        with open(os.path.join(output_loc, 'input_seg_ids.pkl'), 'wb') as f:
            pickle.dump(input_seg_ids, f)
        with open(os.path.join(output_loc, 'labels.pkl'), 'wb') as f:
            pickle.dump(labels, f)

    print('preprocessing')
    input_ids = pad_sequences(input_tokens, maxlen=511, dtype="long", value=0, truncating="pre", padding="post")
    seg_ids = pad_sequences(input_seg_ids, maxlen=511, dtype="long", value=1, truncating="pre", padding="post")
    attention_masks = [[int(token_id > 0) for token_id in sent] for sent in input_ids]


    train_input_ids, validation_input_ids, train_seg_ids, validation_seg_ids, train_attention_masks, validation_attention_masks, train_labels, validation_labels = train_test_split(input_ids, seg_ids, attention_masks, labels, random_state=2019, test_size=0.1)

    del input_ids
    del seg_ids
    del attention_masks
    del train_seg_ids
    del validation_seg_ids
    del train_attention_masks
    del validation_attention_masks

    print('preprocessing trained')
    input0 = list()
    input1 = list()
    for idx, ina in enumerate(train_input_ids):
        count = 0
        input0_0 = list()
        input1_0 = list()
        for i in ina:
            if(i == sep):
                if(count == 0):
                    input0_0.append(i)
                    input0_0.append(cls)
                count = count + 1
            elif(i == cls):
                pass
            else:
                if(count < 2):
                    if(count < 1):
                        input0_0.append(i)
                    else:
                        input1_0.append(i)

        input1_0.append(sep)
        input1_0.append(cls)
        input0.append(input0_0)
        input1.append(input1_0)

    input_ids0 = pad_sequences(input0, maxlen=256, dtype="long", value=0, truncating="pre", padding="pre")
    input_ids1 = pad_sequences(input1, maxlen=256, dtype="long", value=1, truncating="post", padding="post")
    attention_masks = [[int(token_id > 0) for token_id in sent] for sent in input_ids0]

    del input0
    del input1

    print('preprocessing val')
    input0 = list()
    input1 = list()
    for idx, ina in enumerate(validation_input_ids):
        count = 0
        input0_0 = list()
        input1_0 = list()
        for i in ina:
            if(i == sep):
                if(count == 0):
                    input0_0.append(i)
                    input0_0.append(cls)
                count = count + 1
            elif(i == cls):
                pass
            else:
                if(count < 2):
                    if(count < 1):
                        input0_0.append(i)
                    else:
                        input1_0.append(i)

        input1_0.append(sep)
        input1_0.append(cls)
        input0.append(input0_0)
        input1.append(input1_0)

    val_input_ids0 = pad_sequences(input0, maxlen=256, dtype="long", value=0, truncating="pre", padding="pre")
    val_input_ids1 = pad_sequences(input1, maxlen=256, dtype="long", value=1, truncating="post", padding="post")
    val_attention_masks = [[int(token_id > 0) for token_id in sent] for sent in val_input_ids0]

    del input0
    del input1

    input_ids0 = np.array(input_ids0)
    input_ids1 = np.array(input_ids1)
    attention_masks = np.array(attention_masks)
    train_labels = np.array(train_labels)

    val_input_ids0 = np.array(val_input_ids0)
    val_input_ids1 = np.array(val_input_ids1)
    val_attention_masks = np.array(val_attention_masks)
    val_labels = np.array(validation_labels)

    mirrored_strategy = tf.distribute.MirroredStrategy()

    with mirrored_strategy.scope():
        model = concat_model()

    # Include the epoch in the file name (uses `str.format`)
#     checkpoint_path = "./tripletloss/training_model/cp-{epoch:04d}"
    checkpoint_path = "./model/training_model12/cp-{epoch:04d}"
    checkpoint_dir = os.path.dirname(checkpoint_path)

    # Create a callback that saves the model's weights every epoch
    cp_callback = tf.keras.callbacks.ModelCheckpoint(
        filepath=checkpoint_path,
        verbose=1,
        save_freq="epoch")

#     log_dir = "./tripletloss/logs/fit/" + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
    log_dir = "./model/logs12/fit/" + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
    tensorboard_callback = tf.keras.callbacks.TensorBoard(log_dir=log_dir, histogram_freq=1, profile_batch=0)
    # %load_ext tensorboard
    # %tensorboard --logdir logs/fit

    batch_size = 16
    model.fit(x=[input_ids0, input_ids1, attention_masks], y=train_labels,
              validation_data=([val_input_ids0, val_input_ids1, val_attention_masks], val_labels),
              batch_size=batch_size,epochs=16,verbose=1,
              callbacks=[tensorboard_callback, cp_callback])
