import pandas as pd
import lxml.etree as ET
import random
import os

import tensorflow as tf
import tensorflow_addons as tfa
import numpy as np
import pickle
import multiprocessing
from functools import partial

from tensorflow.keras.preprocessing.sequence import pad_sequences

from transformers import XLNetTokenizer
from tqdm import tqdm

def score_book(xlnet_tok_dir, pred_scores_dir, xlnetSeqC, cls, sep, book_id):
    with open(os.path.join(xlnet_tok_dir, book_id + '.pkl'), 'rb') as f:
        d = pickle.load(f)

    m = max(d.keys())

    in_ids_0 = list()
    in_ids_1 = list()
    att = list()

    scores = dict()
    for idx in range(0, m - 1):
        toks1 = d[idx]
        toks2 = d[idx + 1]

        i = idx - 1
        while len(toks1) < 254 and i >= 0:
            toks1 = d[i] + toks1
            i -= 1
        toks1 = toks1[-254:]

        i = idx + 2
        while len(toks2) < 254 and i <= m:
            toks2 = toks2 + d[i]
            i += 1
        toks2 = toks2[:254]


        ids1 = toks1 + [sep] + [cls]
        ids2 = toks2 + [sep] + [cls]


        in_ids_0.append(ids1)
        in_ids_1.append(ids2)

    indexed_tokens_0 = pad_sequences(in_ids_0, maxlen=256, dtype='long', value=0, truncating="pre", padding="pre")
    indexed_tokens_1 = pad_sequences(in_ids_1, maxlen=256, dtype='long', value=1, truncating="post", padding="post")
    attention_masks = [[int(token_id > 0) for token_id in sent] for sent in indexed_tokens_0]

    in_ids_0 = np.array(indexed_tokens_0)
    in_ids_1 = np.array(indexed_tokens_1)
    att = np.array(attention_masks)

    prediction = xlnetSeqC.predict(x=[in_ids_0, in_ids_1, att])

    scores = prediction

    with open(os.path.join(pred_scores_dir, book_id + '.pkl'), 'wb') as f:
        pickle.dump(scores, f)

    return

if __name__ == '__main__':
    # Use appropriate locations
    test_books_list_file = './test_books.txt'

    xlnet_tok_dir = './dataset_new/test_books_xlnet_tok_new/'
    pred_scores_dir = './result/model15/test_preds_06/'
    model_dir = './model/training_model15/cp-0006'

    with open(test_books_list_file, 'r') as f:
        test_book_ids = [x.strip() for x in f.readlines()]

    #  partition = int(sys.argv[1])
    #  from_idx = partition * 1000
    #  to_idx = (partition + 1) * 1000

    #  test_book_ids = test_book_ids[from_idx:to_idx]
    print(len(test_book_ids), 'books')

    tokenizer = XLNetTokenizer.from_pretrained('xlnet-base-cased')

    cls, sep = tokenizer.convert_tokens_to_ids(["<cls>", "<sep>"])

    print('Loading XLNetModel')

    mirrored_strategy = tf.distribute.MirroredStrategy()

    with mirrored_strategy.scope():
        xlnetModel = tf.keras.models.load_model(model_dir)


    for book_id in tqdm(test_book_ids):
        try:
            if os.path.exists(os.path.join(pred_scores_dir, book_id + '.pkl')):
                pass
            else:
                score_book(xlnet_tok_dir, pred_scores_dir, xlnetModel, cls, sep, book_id)
        except Exception as e:
            print(book_id, e)

    print('Done!')
