import pandas as pd
import lxml.etree as ET
import random
import os

import numpy as np
import pickle
import multiprocessing
from functools import partial

from transformers import BertTokenizer
from tqdm.auto import tqdm

def getPrecision(tp, tn, fp, fn):
    return tp / (tp + fp)

def getRecall(tp, tn, fp, fn):
    return tp / (tp + fn)

def getf1(tp, tn, fp, fn):
    pr = getPrecision(tp, tn, fp, fn)
    rc = getRecall(tp, tn, fp, fn)
    return 2 * ( pr * rc) / (pr + rc)

if __name__ == '__main__':
    # Use appropriate locations
    test_books_list_file = './test_books.txt'
    gt_loc = './dataset_new/test_gt.csv'
    prob_loc = './result/model12/test_preds_15/'
    save_score_dir = './result/model12/score_15/'
    with open(test_books_list_file, 'r') as f:
        train_book_ids = [x.strip() for x in f.readlines()]


    print(len(train_book_ids), 'books')

    # Use appropriate locations
    df = pd.read_csv(gt_loc, usecols=['book_id', 'start_para_nums', 'start_para_total'])
    indexdf = df.index
                            #           Predicted           #
    tp = 0                  #   Negative   # #   Postitve   #  # # # # #
    fp = 0     #   Nagative #      tn       #       fp      #   Actual
    fn = 0     #   Positive #      fn       #        tp     #  # # # # #
    tn = 0
    totalPara = 0
    f1_list = list()
    precision_list = list()
    recall_list = list()
    for idx, bookid in enumerate((df['book_id'])):
        if str(bookid) in train_book_ids:
            save_books_score = list()
            try:
                with open(os.path.join(prob_loc, str(bookid) + '.pkl'), 'rb') as f:

                    scorebook = pickle.load(f)
                    bookid_indices = indexdf[df["book_id"] == bookid]
                    start_paras = eval(df.iloc[bookid_indices[0]]['start_para_nums'])

                    booktp = 0
                    bookfn = 0
                    bookfp = 0
                    booktn = 0
                    for score in range(0, len(scorebook)):

                        if int(score) + 1 in start_paras:
                            if scorebook[score] < 0.5:
                                # print('score: ', score)
                                booktp = booktp + 1
                                tp = tp + 1
                                save_books_score.append([int(score) +1, 1, 0, 0, 0])
                            else:
                                bookfn = bookfn + 1
                                fn = fn + 1
                                save_books_score.append([int(score) +1, 0, 1, 0, 0])
                        else:
                            if scorebook[score] < 0.5:
                                bookfp = bookfp + 1
                                fp = fp + 1
                                save_books_score.append([int(score) +1, 0, 0, 1, 0])
                            else:
                                booktn = booktn + 1
                                tn = tn + 1
                                save_books_score.append([int(score) +1, 0, 0, 0, 1])

                    #except:

                    totalPara = totalPara + len(scorebook) - 1




                    f1_list.append(getf1(booktp, booktn, bookfp, bookfn))
                    precision_list.append(getPrecision(tp, tn, fp, fn))
                    recall_list.append(getRecall(tp, tn, fp, fn))

                    df1 = pd.DataFrame(save_books_score)
                    df1.rename(columns={0:'para_num', 1:'tp', 2:'fn', 3:'fp', 4:'tn'}, inplace=True)
                    df1.to_csv(save_score_dir + str(bookid) + '.csv', index=False)
            except:
                print('no book @ : ', str(bookid))
        else:
            print(str(bookid) + ': ' + (bookid in train_book_ids))

    print( 'ParaLen', totalPara)
    print('tp', tp)
    print('fn', fn)
    print('fp', fp)
    print('tn', tn)
    f1 = getf1(tp, tn, fp, fn)
    pr = getPrecision(tp, tn, fp, fn)
    rc = getRecall(tp, tn, fp, fn)
    print(f1, pr, rc)

    with open(os.path.join('./result/model12/', 'accuracyScore_15.txt'), 'w') as f:
        f.write( 'ParaLen\t' + str(totalPara) + '\n')
        f.write( 'tp\t' +  str(tp) + '\n')
        f.write( 'fn\t' +  str(fn) + '\n')
        f.write( 'fp\t' +  str(fp) + '\n')
        f.write( 'tn\t' +  str(tn) + '\n')
        f.write( 'pr\t' +  str(pr) + '\n')
        f.write( 'rc\t' +  str(rc) + '\n')
        f.write( 'f1\t' +  str(f1) + '\n')
